/***********************************************************************************************************************
 *
 * BlueBook - use live code examples in your book
 * Copyright (C) 2011-2011 by Tidalwave s.a.s.
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebook.java.net
 * SCM: https://hg.java.net/hg/bluebook~src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebook.plugin.maven;

import javax.annotation.Nonnull;
import lombok.ToString;
import lombok.Cleanup;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import it.tidalwave.bluebook.filter.java.JavaFilter;
import java.util.Set;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.artifact.InvalidDependencyVersionException;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: BlueBookMojo.java,v 7b7ebf60602a 2011/03/06 21:09:12 fabrizio $
 *
 * @goal generate-filtered-source
 * @threadSafe
 *
 **********************************************************************************************************************/
@ToString
public class BlueBookMojo extends AbstractMojo
  {
    /*******************************************************************************************************************
     *
     * @parameter 
     * @required
     * @readonly
     * 
     ******************************************************************************************************************/
    private String groupId;
    
    /*******************************************************************************************************************
     *
     * @parameter
     * @required
     * @readonly
     * 
     ******************************************************************************************************************/
    private String artifactId;
    
    /*******************************************************************************************************************
     *
     * @parameter
     * @readonly
     * 
     ******************************************************************************************************************/
    private String version;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="sources"
     * @readonly
     * 
     ******************************************************************************************************************/
    private String classifier;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="jar"
     * @readonly
     * 
     ******************************************************************************************************************/
    private String type;
    
    /*******************************************************************************************************************
     *
     * @parameter
     * @required
     * @readonly
     * 
     ******************************************************************************************************************/
    private String fileName;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="true"
     * @readonly
     * 
     ******************************************************************************************************************/
    private boolean skipHeader;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="true"
     * @readonly
     * 
     ******************************************************************************************************************/
    private boolean skipBlockComments;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="false"
     * @readonly
     * 
     ******************************************************************************************************************/
    private boolean skipImports;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="false"
     * @readonly
     * 
     ******************************************************************************************************************/
    private boolean skipPackage;
    
    /*******************************************************************************************************************
     *
     * @parameter 
     * @readonly
     * 
     ******************************************************************************************************************/
    private String includedFragment;

    /*******************************************************************************************************************
     *
     * @parameter 
     * @readonly
     * 
     ******************************************************************************************************************/
    private String excludedFragment;

    /*******************************************************************************************************************
     *
     * @parameter
     * @readonly
     * 
     ******************************************************************************************************************/
    private String outputFile;
    
    /*******************************************************************************************************************
     *
     * @parameter expression=
     *  "${component.org.apache.maven.artifact.factory.ArtifactFactory}"
     * @required
     * @readonly
     * 
     ******************************************************************************************************************/
    protected ArtifactFactory factory;

    /*******************************************************************************************************************
     *
     * @parameter expression="${component.org.apache.maven.artifact.resolver.ArtifactResolver}"
     * @required
     * @readonly
     * 
     ******************************************************************************************************************/
    protected ArtifactResolver artifactResolver;

    /*******************************************************************************************************************
     *
     * List of Remote Repositories used by the resolver
     * 
     * @parameter expression="${project.remoteArtifactRepositories}"
     * @readonly
     * @required
     * 
     ******************************************************************************************************************/
    protected List remoteRepositories;

    /*******************************************************************************************************************
     *
     * Location of the local repository.
     * 
     * @parameter expression="${localRepository}"
     * @readonly
     * @required
     * 
     ******************************************************************************************************************/
    protected ArtifactRepository localRepository;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="${project}"
     * 
     ******************************************************************************************************************/
    protected MavenProject mavenProject;
    
    /*******************************************************************************************************************
     *
     * @parameter default-value="${project.artifactMetadataSource}"
     * 
     ******************************************************************************************************************/
    protected ArtifactMetadataSource artifactMetadataSource;
    
    /*******************************************************************************************************************
     *
     * TODO: extract an interface, make it a mojo parameter
     * 
     ******************************************************************************************************************/
    protected ProgramListingTemplate template = new ProgramListingTemplate();
    
    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    @Override
    public void execute() 
      throws MojoExecutionException
      {
        try 
          {
            final Log log = getLog();
            log.info("Creating generated example snippets...");
            log.debug("mojo: " + this);
           
            if (version == null)
              {
                searchVersionInDependencyManagement();
              }
            
            if (version == null)
              {
                throw new MojoExecutionException("Can't find version of " + groupId + ":" + artifactId + " in dependencyManagement");  
              }
            
            final Artifact artifact = factory.createArtifactWithClassifier(groupId, artifactId, version, type, classifier);
            artifactResolver.resolve(artifact, remoteRepositories, localRepository);
                        
            final File artifactFile = artifact.getFile();
            log.info("file: " + artifactFile);
            final JarFile jarFile = new JarFile(artifactFile);
            final JarEntry jarEntry = jarFile.getJarEntry(fileName);
            
            if (jarEntry == null)
              {
                throw new MojoExecutionException("Cannot find " + fileName + " in " + artifact);  
                // TODO: see http://java.net/jira/browse/BLUEBOOK-6
              }
            
            final File targetFile = new File("target/generated-sources/docbkx/" + outputFile);
            log.info("Creating " + outputFile + "...");
            targetFile.getParentFile().mkdirs();
            
            process(jarFile, jarEntry, targetFile);
          }
        catch (InvalidDependencyVersionException e) 
          {
            throw new MojoExecutionException("", e);
          } 
//        catch (ProjectBuildingException e) 
//          {
//            throw new MojoExecutionException("", e);
//          }
        catch (IOException e) 
          {
            throw new MojoExecutionException("", e);
          }
        catch (ArtifactResolutionException e) 
          {
            throw new MojoExecutionException("", e);
          }
        catch (ArtifactNotFoundException e) 
          {
            throw new MojoExecutionException("", e);
          }
      }

    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private void searchVersionInDependencyManagement() 
      throws InvalidDependencyVersionException, ArtifactResolutionException, ArtifactNotFoundException
      {
        final Log log = getLog();
        final Set<?> dependencyArtifacts = mavenProject.createArtifacts(factory, null, null);
        final ArtifactResolutionResult result = artifactResolver.resolveTransitively(dependencyArtifacts, 
                                                                                     mavenProject.getArtifact(), 
                                                                                     remoteRepositories, 
                                                                                     localRepository, 
                                                                                     artifactMetadataSource);
        for (final Artifact depManArtifact : (Set<Artifact>)result.getArtifacts()) 
          {
            // do not use classifier and type in the comparison: so we can take advantage of the regular
            // default dependency (i.e. the jar) insteading of requiring one with specific classifier.
            //
            if (depManArtifact.getGroupId().equals(groupId) 
                && depManArtifact.getArtifactId().equals(artifactId))
              {
                version = depManArtifact.getVersion();
                log.info("Artifact version from dependencyManagement: " + version);
                break;
              }
          }
      }

    /*******************************************************************************************************************
     *
     * 
     ******************************************************************************************************************/
    private void process (final @Nonnull JarFile jarFile, final @Nonnull JarEntry jarEntry, @Nonnull final File targetFile) 
      throws IOException 
      {        
        final @Cleanup Reader r = new InputStreamReader(jarFile.getInputStream(jarEntry));
        final @Cleanup FileWriter w = new FileWriter(targetFile);
        
        final JavaFilter javaFilter = new JavaFilter();
        javaFilter.setSkipBlockComments(skipBlockComments);
        javaFilter.setSkipHeader(skipHeader);
        javaFilter.setSkipImports(skipImports);
        javaFilter.setSkipPackage(skipPackage);
        javaFilter.setIncludedFragment(includedFragment);
        javaFilter.setExcludedFragment(excludedFragment);
        
        template.process(javaFilter, r, w);
      }
  }
